# Copyright 2005-2007, Google, Inc.

# This software is provided as an example of using the Ruby interface
# to SketchUp.

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------

# Name: Construction Line Tool
#
require 'sketchup.rb'
#require "langhandler.rb"

#$exStrings = LanguageHandler.new("CLineTool")

#-----------------------------------------------------------------------------

# This example shows the implementation of a simple line tool.  This tool
# is similar to the Pencil tool in SketchUp, but it create construction lines
# rather than edges
#
# HISTORY
#
# 2011-01-04 <jim.foltz@gmail.com>
#   * Added enableVCB? method (TT)
#   * Added false after calling onSetCursor in the onKeyDown method. 
#     - TT - must do something after calling onSetCursor in onKeyDown or onLButton
#
# Earlier
#   * Tab key cycles stipple type insead of showing dialog.
#

module JF
    class CLineTool

        def initialize

            # Modified by Henry Schorradt
            @p1vorhanden = 0;
            @p2vorhanden = 0;


            # This is the standard Ruby initialize method that is called when you create
            # a new object.
            @ip1 = nil
            @ip2 = nil
            @xdown = 0
            @ydown = 0
            #@draw_cpoints = true
            c = File.join(File.dirname(__FILE__), "clinetool_cursor.png")
            ltcursor = UI::create_cursor(c, 1, 30)
            c = File.join(File.dirname(__FILE__), "clinetool_cursor_p.png")
            ltcursor_p = UI::create_cursor(c, 1, 30)
            c = File.join(File.dirname(__FILE__), "clinetool_cursor_i.png")
            ltcursor_i = UI::create_cursor(c, 1, 30)
            @cursors = [ltcursor, ltcursor_p, ltcursor_i]
            @opts = {}
            load_opts()
        end

        def onSetCursor
            #if @draw_cpoints
            #  UI::set_cursor(@ltcursor_p)
            #else
            #  UI::set_cursor(@ltcursor)
            #end
            #case @opts['mode']
            #when 0
            #    UI::set_cursor(@ltcursor)
            #when 1
            #    UI::set_cursor(@ltcursor_p)
            #when 2
            #    UI::set_cursor(@ltcursor_i)
            #end
            UI.set_cursor(@cursors[@opts['mode']])
            #Sketchup.active_model.active_view.refresh
        end

        # The activate method is called by SketchUp when the tool is first selected.
        # it is a good place to put most of your initialization
        def activate
            # The Sketchup::InputPoint class is used to get 3D points from screen
            # positions.  It uses the SketchUp inferencing code.
            # In this tool, we will have two points for the endpoints of the line.
            @ip         = Sketchup::InputPoint.new
            @ip1        = Sketchup::InputPoint.new
            @ip2        = Sketchup::InputPoint.new
            @drawn      = false
            @last_drawn = nil
            @stipple    = ["_",  ".",  "-", "-.-"]

            Sketchup::set_status_text("Length", SB_VCB_LABEL)

            # Unhide Construction Geometry
            if Sketchup.active_model.rendering_options["HideConstructionGeometry"] == true 
                Sketchup.active_model.rendering_options["HideConstructionGeometry"] = false 
            end

            self.reset(nil)
        end

        def resume(view)
            txt = "Construction Line Tool: CTRL = cycle mode. TAB = set stipple."
            Sketchup.set_status_text(txt, SB_PROMPT)
        end


        # deactivate is called when the tool is deactivated because
        # a different tool was selected
        def deactivate(view)
            view.invalidate if @drawn
            save_opts()
        end

        # The onMouseMove method is called whenever the user moves the mouse.
        # because it is called so often, it is important to try to make it efficient.
        # In a lot of tools, your main interaction will occur in this method.
        def onMouseMove(flags, x, y, view)
            if( @state == 0 )
                # We are getting the first end of the line.  Call the pick method
                # on the InputPoint to get a 3D position from the 2D screen position
                # that is bassed as an argument to this method.
                @ip.pick view, x, y
                if( @ip != @ip1 )
                    # if the point has changed from the last one we got, then
                    # see if we need to display the point.  We need to display it
                    # if it has a display representation or if the previous point
                    # was displayed.  The invalidate method on the view is used
                    # to tell the view that something has changed so that you need
                    # to refresh the view.
                    view.invalidate if( @ip.display? or @ip1.display? )
                    @ip1.copy! @ip

                    # set the tooltip that should be displayed to this point
                    view.tooltip = @ip1.tooltip
                end
            else
                # Getting the second end of the line
                # If you pass in another InputPoint on the pick method of InputPoint
                # it uses that second point to do additional inferencing such as
                # parallel to an axis.
                @ip2.pick view, x, y, @ip1
                view.tooltip = @ip2.tooltip if( @ip2.valid? )

                # Update the length displayed in the VCB
                if( @ip2.valid? )
                    length = @ip1.position.distance(@ip2.position)
                    Sketchup::set_status_text length.to_s, SB_VCB_VALUE
                end

                # Check to see if the mouse was moved far enough to create a line.
                # This is used so that you can create a line by either draggin
                # or doing click-move-click
                if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
                    @dragging = true
                end
                view.invalidate
            end
        end

        # The onLButtonDOwn method is called when the user presses the left mouse button.
        def onLButtonDown(flags, x, y, view)
            # When the user clicks the first time, we switch to getting the
            # second point.  When they click a second time we create the line
            if( @state == 0 )
                @ip1.pick view, x, y
                if( @ip1.valid? )
                    @state = 1
                    txt = "Select second end. Ctrl = toggle points "
                    if @opts['mode']==1#@draw_cpoints
                        txt << "off."
                    else
                        txt << "on."
                    end
                    txt << "TAB = stipple."
                    Sketchup::set_status_text(txt, SB_PROMPT)
                    @xdown = x
                    @ydown = y
                end
            else
                # create the line on the second click
                if( @ip2.valid? )
                    self.create_geometry(@ip1.position, @ip2.position,view)
                    self.reset(view)
                end
            end

            # Clear any inference lock
            view.lock_inference
        end

        # The onLButtonUp method is called when the user releases the left mouse button.
        def onLButtonUp(flags, x, y, view)
            # If we are doing a drag, then create the line on the mouse up event
            if( @dragging && @ip2.valid? )
                self.create_geometry(@ip1.position, @ip2.position,view)
                self.reset(view)
            end
        end

        # onKeyDown is called when the user presses a key on the keyboard.
        # We are checking it here to see if the user pressed the shift key
        # so that we can do inference locking
        def onKeyDown(key, repeat, flags, view)
            #puts :onKeyDown
            if key == 9
                #prompts = ["Stipple"]
                #defaults = [ @opts['stipple'] ]
                #choices = ["_|.|-|-.-"]
                #ret = UI.inputbox(prompts, defaults, choices)
                #return if ret == false
                @stipple.push(@stipple.shift)
                @opts['stipple'] = @stipple[0]
                #Sketchup.vcb_label = "Stipple"
                #Sketchup.vcb_value = @opts['stipple']
                #Sketchup.status_text = "Stipple: " + @opts['stipple']
                view.invalidate
            end
            if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
                @shift_down_time = Time.now

                # if we already have an inference lock, then unlock it
                if( view.inference_locked? )
                    # calling lock_inference with no arguments actually unlocks
                    view.lock_inference
                elsif( @state == 0 && @ip1.valid? )
                    view.lock_inference @ip1
                    view.line_width = 3
                elsif( @state == 1 && @ip2.valid? )
                    view.lock_inference @ip2, @ip1
                    view.line_width = 3
                end
            end
            if( key == COPY_MODIFIER_KEY )
                #@draw_cpoints = !@draw_cpoints
                inc_mode()
                if @state == 0
                    txt = "Select first end. Ctrl = toggle points "
                else
                    txt = "Select second end. Ctrl = toggle points "
                end
                if @opts['mode'] == 1#@draw_cpoints
                    txt << "off."
                else
                    txt << "on."
                end
                Sketchup::set_status_text(txt, SB_PROMPT)
            end
            #view.refresh
            self.onSetCursor() # forces change of cursor on screen
            #onSetCursor
            # need to do something after onSetCursor
            false # TT - The culprit is onSetCursor when it's called last in onKey or onLButton events. 
        end

        def inc_mode
            @opts['mode'] += 1
            if @opts['mode'] > 2
                @opts['mode'] = 0
            end
        end

        # onKeyUp is called when the user releases the key
        # We use this to unlock the interence
        # If the user holds down the shift key for more than 1/2 second, then we
        # unlock the inference on the release.  Otherwise, the user presses shift
        # once to lock and a second time to unlock.
        def onKeyUp(key, repeat, flags, view)
            if( key == CONSTRAIN_MODIFIER_KEY && view.inference_locked? && (Time.now - @shift_down_time) > 0.5 )
                view.lock_inference
            end
            #onSetCursor # trigger change of cursore on screen
        end

        # onUserText is called when the user enters something into the VCB
        # In this implementation, we create a line of the entered length if
        # the user types a length while selecting the second point
        def onUserText(text, view)
            # We only accept input when the state is 1 (i.e. getting the second point)
            # This could be enhanced to also modify the last line created if a length
            # is entered after creating a line.
            #return if not @state == 1
            #return if not @ip2.valid?
            #p @ip2.valid?

            # The user may type in something that we can't parse as a length
            # so we set up some exception handling to trap that
            begin
                value = text.to_l
            rescue
                # Error parsing the text
                UI.beep
                puts "Cannot convert #{text} to a Length"
                value = nil
                Sketchup::set_status_text "", SB_VCB_VALUE
            end
            return if !value

            if @state == 1 and @ip2.valid?
                # Compute the direction and the second point
                pt1 = @ip1.position
                vec = @ip2.position - pt1
                if( vec.length == 0.0 )
                    UI.beep
                    return
                end
                vec.length = value
                pt2 = pt1 + vec

                # Create a line
                self.create_geometry(pt1, pt2, view)
                self.reset(view)
            end
            if @last_drawn and @state == 0
                pt1 = @last_drawn[0].start
                vec = @last_drawn[0].end - @last_drawn[0].start
                vec.length = value
                pt2 = pt1 + vec
                view.model.active_entities.erase_entities @last_drawn
                #@last_drawn = nil
                self.create_geometry(pt1, pt2, view)
                self.reset view
            end
        end #onUserText

        # The draw method is called whenever the view is refreshed.  It lets the
        # tool draw any temporary geometry that it needs to.
        def draw(view)
            if( @ip1.valid? )
                if( @ip1.display? )
                    @ip1.draw(view)
                    @drawn = true
                end

                if( @ip2.valid? )
                    @ip2.draw(view) if( @ip2.display? )

                    # The set_color_from_line method determines what color
                    # to use to draw a line based on its direction.  For example
                    # red, green or blue.
                    if view.inference_locked?
                        line_width = 2
                    else
                        line_width = 1
                    end
                    view.line_width = line_width
                    view.line_stipple = @opts['stipple']
                    view.set_color_from_line(@ip1, @ip2)
                    self.draw_geometry(@ip1.position, @ip2.position, view)
                    @drawn = true
                end
            end
        end

        # onCancel is called when the user hits the escape key
        def onCancel(flag, view)
            self.reset(view)
        end


        # The following methods are not directly called from SketchUp.  They are
        # internal methods that are used to support the other methods in this class.

        # Reset the tool back to its initial state
        def reset(view)
            # This variable keeps track of which point we are currently getting
            @state = 0

            # Display a prompt on the status bar
            txt = "Select first end. Ctrl = toggle points "
            if @opts['mode'] == 1#@draw_cpoints
                txt << "off."
            else
                txt << "on."
            end
            txt << "TAB = stipple."
            Sketchup::set_status_text(txt, SB_PROMPT)

            # clear the InputPoints
            @ip1.clear
            @ip2.clear
            @ip.clear

            if( view )
                view.tooltip = nil
                view.invalidate if @drawn
            end

            @drawn = false
            @dragging = false
        end

        # Create new geometry when the user has selected two points.
        def create_geometry(p1, p2, view)
            Sketchup.active_model.start_operation("CLineTool")
            @layer = Sketchup.active_model.layers.add("CLineTool")
            if @opts['mode'] == 0
                cline = view.model.active_entities.add_cline(p1,p2, @opts['stipple'])
                cline.layer = @layer
                @last_drawn = [cline]
            elsif @opts['mode'] == 1 # @draw_cpoints
                # Modified by Henry Schorradt
                cline = view.model.active_entities.add_cline(p1,p2, @opts['stipple'])
                cline.layer = @layer
                @last_drawn = [cline]
                ent = Sketchup.active_model.active_entities
                ent.each {|e|
                    if (e.typename == "ConstructionPoint")
                        if( e.position == p1 )
                            @p1vorhanden = 1
                        end
                        if( e.position == p2 )
                            @p2vorhanden = 1 
                        end
                    end
                }
                # Modified by Henry Schorradt
                if ( @p1vorhanden == 0)
                    cpoint =  view.model.active_entities.add_cpoint(p1)
                    cpoint.layer = @layer
                    #@last_drawn << view.model.active_entities.add_cpoint(p1)
                    @last_drawn << cpoint
                end
                if ( @p2vorhanden == 0)
                    cpoint =  view.model.active_entities.add_cpoint(p2)
                    cpoint.layer = @layer
                    #@last_drawn << view.model.active_entities.add_cpoint(p2)
                    @last_drawn << cpoint
                end
            elsif @opts['mode'] == 2
                vec = p1 - p2
                cline = view.model.active_entities.add_cline(p1, p1-p2, @opts['stipple'])
                cline.layer = @layer
                #@last_drawn = [cline]
            end
            #p @last_drawn

            # Modified by Henry Schorradt
            @p1vorhanden = 0;
            @p2vorhanden = 0;

            Sketchup.active_model.commit_operation
        end

        # Draw the geometry
        def draw_geometry(pt1, pt2, view)
            view.draw_line(pt1, pt2)
        end

        def getExtents
            bbox=Sketchup.active_model.bounds
            bbox.add(@ip.position) if @ip and @ip.valid?
            bbox.add(@ip1.position)if @ip1 and @ip1.valid?
            bbox.add(@ip2.position)if @ip2 and @ip2.valid?
            #bbox.add(@ip3.position)if @ip3 and @ip3.valid?
            ### etc
            return bbox
        end

        def load_opts
            key = "JF::CLineTool"
            opts = Sketchup.read_default(key, 'options')
            @opts ||= {}
            @opts['mode'] = 1
            @opts['stipple'] = '_'
            if opts.is_a? Array
                opts.each {|o|
                    k, v = o
                    @opts[k] = v
                }
            end
        end

        def save_opts
            key = "JF::CLineTool"
            o = @opts.inspect.to_s.gsub(/"/, "'")
            Sketchup.write_default(key, 'options', @opts.to_a)
        end

        def enableVCB?
            true
        end

    end # class LineTool

end # module JF

#-----------------------------------------------------------------------------
# This functions is just a shortcut for selecting the new tool
#def linetool
#Sketchup.active_model.select_tool LineTool.new
#end

unless file_loaded?(__FILE__)
    cmd = UI::Command.new("Construction Line Tool") {Sketchup.active_model.select_tool(JF::CLineTool.new)}
    my_dir         = File.dirname(File.expand_path(__FILE__))
    cmd.small_icon = File.join(my_dir, "clinetool_icon_sm.png")
    cmd.large_icon = File.join(my_dir, "clinetool_icon_lg.png")
    cmd.tooltip    = "Draw Construction Lines"
    cmd.menu_text  = "Construction Lines"

    menu = UI.menu("Draw")
    menu.add_item(cmd)

    tb = UI.toolbar("Construction Line Tool")
    tb.add_item(cmd)
    if tb.get_last_state == TB_VISIBLE
        UI.start_timer(0.1, false) { tb.restore }
    elsif tb.get_last_state == TB_NEVER_SHOWN
        tb.show
    end
end
file_loaded(__FILE__)
